#include "dune-raw-data/HDFUtils/HDFFileReader.hh"
#include "dune-raw-data/Overlays/FragmentType.hh"
#include "dune-raw-data/Overlays/FelixFragment.hh"
#include "dune-raw-data/Overlays/SSPFragment.hh"
#include "artdaq-core/Utilities/TimeUtils.hh"
#include <boost/algorithm/string.hpp>
#include <regex>
#include <set>
#include <sstream>

#include <sys/time.h>
#include <sys/resource.h>

namespace dune {
  namespace HDFFileReader {

    // *****************************************
    // *** General-purpose data and routines ***
    // *****************************************

    HDFFileInfoPtr openFile(const std::string& fileName)
    {
      HDFFileInfoPtr hdfFileInfoPtr(new HDFFileInfo());
      hdfFileInfoPtr->filePtr = new H5::H5File(fileName, H5F_ACC_RDONLY);
      hdfFileInfoPtr->bytesWritten = 0;
      hdfFileInfoPtr->fileName = fileName;
      if (hdfFileInfoPtr->filePtr->attrExists("attribute_style_version"))
      {
	float versionNumber;
	H5::Attribute versionAttr = hdfFileInfoPtr->filePtr->openAttribute("attribute_style_version");
	versionAttr.read(versionAttr.getDataType(), &versionNumber);
	if (versionNumber >= 0.8)
        {
	  H5::Attribute runAttr = hdfFileInfoPtr->filePtr->openAttribute("run_number");
	  runAttr.read(runAttr.getDataType(), &hdfFileInfoPtr->runNumber);
	}
      }
      return hdfFileInfoPtr;
    }

    HDFFileInfoPtr reopenFile(HDFFileInfoPtr oldHdfFileInfoPtr)
    {
      oldHdfFileInfoPtr->filePtr->close();
      delete oldHdfFileInfoPtr->filePtr;
      oldHdfFileInfoPtr->filePtr = nullptr;

      HDFFileInfoPtr newHdfFileInfoPtr(new HDFFileInfo());
      newHdfFileInfoPtr->filePtr = new H5::H5File(oldHdfFileInfoPtr->fileName, H5F_ACC_RDONLY);
      newHdfFileInfoPtr->bytesWritten = 0;
      newHdfFileInfoPtr->fileName = oldHdfFileInfoPtr->fileName;
      newHdfFileInfoPtr->runNumber = oldHdfFileInfoPtr->runNumber;
      return newHdfFileInfoPtr;
    }

    void closeFile(HDFFileInfoPtr hdfFileInfoPtr)
    {
      hdfFileInfoPtr->filePtr->close();
      delete hdfFileInfoPtr->filePtr;
      hdfFileInfoPtr->filePtr = nullptr;
    }

    std::list<std::string> getTopLevelGroupNames(HDFFileInfoPtr& hdfFileInfoPtr)
    {
      std::list<std::string> theList;
      for (size_t idx = 0; idx < hdfFileInfoPtr->filePtr->getNumObjs(); ++idx)
      {
	std::string tlgName = hdfFileInfoPtr->filePtr->getObjnameByIdx(idx);
	theList.push_back(tlgName);
      }
      return theList;
    }

    std::list<std::string> getMidLevelGroupNames(H5::Group& theGroup)
    {
      std::list<std::string> theList;
      for (size_t idx = 0; idx < theGroup.getNumObjs(); ++idx)
      {
	std::string tlgName = theGroup.getObjnameByIdx(idx);
	theList.push_back(tlgName);
      }
      return theList;
    }

    H5::Group getGroupFromPath(HDFFileInfoPtr& hdfFileInfoPtr, const std::string& groupPath)
    {
      std::vector<std::string> pathElements;
      boost::split(pathElements, groupPath, boost::is_any_of("/"));
      if (pathElements.size() == 0 || boost::trim_copy(pathElements[0]).empty())
      {
	throw "Error in HDFFileReader::getGroupFromPath: unable to split path " + groupPath;
      }

      // look up the top-level group
      std::string topLevelGroupName = boost::trim_copy(pathElements[0]);
      H5::Group workingGroup = hdfFileInfoPtr->filePtr->openGroup(topLevelGroupName);

      // create (if needed) and look up the remaining requested groups
      for (size_t idx = 1; idx < pathElements.size(); ++idx)
      {
	std::string childGroupName = boost::trim_copy(pathElements[idx]);
	if (childGroupName.empty())
	{
	  throw "Error in HDFFileReader::getGroupFromPath: child group name is an empty string in path " + groupPath;
	}
	H5::Group childGroup = workingGroup.openGroup(childGroupName);
	workingGroup = childGroup;
      }

      return workingGroup;
    }

    H5::DataSet getDataSetFromPath(HDFFileInfoPtr& hdfFileInfoPtr, const std::string& dataSetPath)
    {
      std::vector<std::string> pathElements;
      boost::split(pathElements, dataSetPath, boost::is_any_of("/"));
      if (pathElements.size() == 0 || boost::trim_copy(pathElements[0]).empty())
      {
	throw "Error in HDFFileReader::getDataSetFromPath: unable to split path " + dataSetPath;
      }

      // check for top-level dataset
      if (pathElements.size() == 1)
      {
	return hdfFileInfoPtr->filePtr->openDataSet(boost::trim_copy(pathElements[0]));
      }

      std::string parentGroupPath = boost::trim_copy(pathElements[0]);
      for (size_t idx = 1; idx < (pathElements.size()-1); ++idx)
      {
	parentGroupPath += "/" + boost::trim_copy(pathElements[idx]);
      }
      H5::Group parentGroup = getGroupFromPath(hdfFileInfoPtr, parentGroupPath);
      return parentGroup.openDataSet(boost::trim_copy(pathElements[ (pathElements.size()-1) ]));
    }


    // ************************************************
    // *** Trigger-Record-related data and routines ***
    // ************************************************

    FragmentListsByType getFragmentsForEvent(HDFFileInfoPtr& hdfFileInfoPtr, const std::string& topLevelGroupName)
    {
      FragmentListsByType fragmentMap;
      artdaq::Fragment::fragment_id_t sampleFragmentID = 1001;

      //struct rusage blah;
      //getrusage(RUSAGE_SELF, &blah);
      //std::cout << "KAB " << __LINE__ << " rusage max resident memory " << blah.ru_maxrss << std::endl;

      H5::Group requestedGroup = getGroupFromPath(hdfFileInfoPtr, topLevelGroupName);
      {
	std::list<std::string> detectorTypeNames = getMidLevelGroupNames(requestedGroup);
	for (auto& detectorTypeName : detectorTypeNames)
	{
 	  std::string subdetGroupPath = topLevelGroupName + "/" + detectorTypeName;
	  H5::Group subdetGroup = getGroupFromPath(hdfFileInfoPtr, subdetGroupPath);
	  {
	    fragmentMap[detectorTypeName] = std::make_unique<artdaq::Fragments>();

	    std::list<std::string> subdetGeoNames = getMidLevelGroupNames(subdetGroup);
	    for (auto& subdetGeoName : subdetGeoNames)
	    {
	      std::string dataSetPath = subdetGroupPath + "/" + subdetGeoName;
	      H5::DataSet theDataSet = getDataSetFromPath(hdfFileInfoPtr, dataSetPath);

	      //std::cout << "Reading DataSet " << dataSetPath << ", sizes " << theDataSet.getStorageSize()
	      //	<< " " << theDataSet.getDataType().getSize() << " " << theDataSet.getElementCount()
	      //	<< std::endl;

	      artdaq::FragmentPtr frag;
	      artdaq::Fragment::type_t type = artdaq::Fragment::InvalidFragmentType;
	      uint32_t payloadOffset = 0;
	      if (detectorTypeName == FELIX_GROUP_NAME)
	      {
		frag = artdaq::Fragment::FragmentBytes(theDataSet.getStorageSize());
		type = dune::detail::FELIX;

		uint64_t windowStart = 0;
		uint64_t windowEnd = 0;
		uint64_t triggerTime = 0;

		if (requestedGroup.attrExists("timeslice_start"))
		{
		  H5::Attribute tsAttr = requestedGroup.openAttribute("timeslice_start");
		  tsAttr.read(tsAttr.getDataType(), &windowStart);
		}
		else
		{
		  if (theDataSet.attrExists("timestamp_of_first_frame"))
		  {
		    H5::Attribute tsAttr = theDataSet.openAttribute("timestamp_of_first_frame");
		    tsAttr.read(tsAttr.getDataType(), &windowStart);
		  }
		}
		if (requestedGroup.attrExists("timeslice_end"))
		{
		  H5::Attribute tsAttr = requestedGroup.openAttribute("timeslice_end");
		  tsAttr.read(tsAttr.getDataType(), &windowEnd);
		}
		else
		{
		  if (theDataSet.attrExists("timestamp_of_last_frame"))
		  {
		    H5::Attribute tsAttr = theDataSet.openAttribute("timestamp_of_last_frame");
		    tsAttr.read(tsAttr.getDataType(), &windowEnd);
		  }
		}
		if (theDataSet.attrExists("trigger_timestamp"))
	        {
		  H5::Attribute tsAttr = theDataSet.openAttribute("trigger_timestamp");
		  tsAttr.read(tsAttr.getDataType(), &triggerTime);
		}
		else
		{
		  if (requestedGroup.attrExists("trigger_timestamp"))
		  {
		    H5::Attribute tsAttr = requestedGroup.openAttribute("trigger_timestamp");
		    tsAttr.read(tsAttr.getDataType(), &triggerTime);
		  }
		  else
		  {
		    if (requestedGroup.attrExists("timeslice_start"))
		    {
		      H5::Attribute tsAttr = requestedGroup.openAttribute("timeslice_start");
		      tsAttr.read(tsAttr.getDataType(), &triggerTime);
		    }
		  }
		}

		dune::FelixFragmentBase::Metadata felixMetadata;
		felixMetadata.control_word = 0xabc;
		felixMetadata.version = 1;
		felixMetadata.reordered = 0;
		felixMetadata.compressed = 0;
		felixMetadata.offset_frames = (uint32_t) ((triggerTime - windowStart) / 25);
		felixMetadata.window_frames = (uint32_t) ((windowEnd - windowStart) / 25);
		if (theDataSet.attrExists("number_of_frames"))
	        {
		  H5::Attribute tsAttr = theDataSet.openAttribute("number_of_frames");
		  tsAttr.read(tsAttr.getDataType(), &felixMetadata.num_frames);
		}
		frag->setMetadata(felixMetadata);
	      }

	      uint64_t pdsPayloadHeaderWord = 0;
	      if (detectorTypeName == PDS_GROUP_NAME)
	      {
		payloadOffset = sizeof(uint64_t);
		frag = artdaq::Fragment::FragmentBytes(theDataSet.getStorageSize() + sizeof(uint64_t));
		type = dune::detail::PHOTON;

		// fill in the SSPFragment metadata, as much as we can
		SSPDAQ::MillisliceHeader pdsMetadata;
		pdsMetadata.triggerTime = 0;
		pdsMetadata.triggerType = 0;
		if (requestedGroup.attrExists("trigger_timestamp"))
		{
		  uint64_t timestamp;
		  H5::Attribute tsAttr = requestedGroup.openAttribute("trigger_timestamp");
		  tsAttr.read(tsAttr.getDataType(), &timestamp);
		  pdsMetadata.triggerTime = timestamp * 3;
		}
		if (requestedGroup.attrExists("timeslice_start"))
		{
		  uint64_t timestamp;
		  H5::Attribute tsAttr = requestedGroup.openAttribute("timeslice_start");
		  tsAttr.read(tsAttr.getDataType(), &timestamp);
		  pdsMetadata.startTime = timestamp * 3;
		  if (pdsMetadata.triggerTime == 0) {pdsMetadata.triggerTime = timestamp * 3;}
		}
		else
		{
		  if (theDataSet.attrExists("timestamp_of_first_packet"))
		  {
		    uint64_t timestamp;
		    H5::Attribute tsAttr = theDataSet.openAttribute("timestamp_of_first_packet");
		    tsAttr.read(tsAttr.getDataType(), &timestamp);
		    pdsMetadata.startTime = timestamp * 3;
		  }
		}
		if (requestedGroup.attrExists("timeslice_end"))
		{
		  uint64_t timestamp;
		  H5::Attribute tsAttr = requestedGroup.openAttribute("timeslice_end");
		  tsAttr.read(tsAttr.getDataType(), &timestamp);
		  pdsMetadata.endTime = timestamp * 3;
		}
		else
		{
		  if (theDataSet.attrExists("timestamp_of_last_packet"))
		  {
		    uint64_t timestamp;
		    H5::Attribute tsAttr = theDataSet.openAttribute("timestamp_of_last_packet");
		    tsAttr.read(tsAttr.getDataType(), &timestamp);
		    pdsMetadata.endTime = timestamp * 3;
		  }
		}
		if (theDataSet.attrExists("number_of_packets"))
		{
		  uint32_t packetCount;
		  H5::Attribute pcAttr = theDataSet.openAttribute("number_of_packets");
		  pcAttr.read(pcAttr.getDataType(), &packetCount);
		  pdsMetadata.nTriggers = packetCount;
		}
		if (theDataSet.attrExists("size_in_bytes"))
		{
		  uint32_t sizeBytes;
		  H5::Attribute pcAttr = theDataSet.openAttribute("size_in_bytes");
		  pcAttr.read(pcAttr.getDataType(), &sizeBytes);
		  pdsMetadata.length = 9 + (sizeBytes / sizeof(uint32_t));
		  pdsPayloadHeaderWord |= ((2 + (sizeBytes / sizeof(uint32_t))) & 0xffffffff);
		}
		else
		{
		  pdsMetadata.length = 9 + (theDataSet.getStorageSize() / sizeof(uint32_t));
		  pdsPayloadHeaderWord |= ((2 + (theDataSet.getStorageSize() / sizeof(uint32_t))) & 0xffffffff);
		}

		frag->setMetadata(pdsMetadata);
		pdsPayloadHeaderWord |= 0x000003e700000000;  // run number always seems to be set to 999
	      }

	      artdaq::Fragment::sequence_id_t seqID;
	      artdaq::Fragment::fragment_id_t fragID;
	      artdaq::Fragment::timestamp_t headerTimestamp;

	      std::stringstream seqStr(topLevelGroupName);
	      seqStr >> seqID;
	      // sequence ID in artdaq::Fragment header is only 48 bits
	      seqID &= 0xffffffffffff;

	      fragID = sampleFragmentID++;

	      if (requestedGroup.attrExists("trigger_timestamp"))
	      {
		uint64_t timestamp;
		H5::Attribute tsAttr = requestedGroup.openAttribute("trigger_timestamp");
		tsAttr.read(tsAttr.getDataType(), &timestamp);
		headerTimestamp = timestamp;
	      }
	      else
	      {
		std::stringstream timeStr(topLevelGroupName);
		timeStr >> headerTimestamp;
	      }

	      frag->setUserType(type);
	      frag->setSequenceID(seqID);
	      frag->setFragmentID(fragID);
	      frag->setTimestamp(headerTimestamp);

	      //std::cout << "Attempting to determine sequence ID " << topLevelGroupName << " " << seqID
	      //          << " " << ((uint64_t)frag->sequenceID()) << std::endl;

	      //TLOG(TLVL_READFRAGMENT_V) << "readFragment_: Reading payload data into Fragment BEGIN";
	      theDataSet.read(frag->dataBeginBytes() + payloadOffset, theDataSet.getDataType());

	      if (detectorTypeName == PDS_GROUP_NAME)
	      {
		memcpy(frag->dataBeginBytes(), &pdsPayloadHeaderWord, sizeof(uint64_t));
	      }

	      fragmentMap[detectorTypeName]->emplace_back(std::move(*(frag.release())));
	    }
	  }
	}
      }

      hdfFileInfoPtr->filePtr->flush(H5F_SCOPE_GLOBAL);
      return fragmentMap;
    }

  } // namespace HDFFileReader
}   // namespace dune
